@@ -47,8 +47,8 @@
 #include "radeon.h"
 #include "radeon_reg.h"
 #include "radeon_macros.h"
-
 #include "radeon_version.h"
+#include "radeon_atombios.h"
 
 #include "xf86Modes.h"
 				/* DDC support */
@@ -59,26 +59,29 @@ void RADEONSetPitch (ScrnInfoPtr pScrn)
 {
     int  dummy = pScrn->virtualX;
     RADEONInfoPtr info = RADEONPTR(pScrn);
+    int pitch_mask = 0;
+    int align_large;
+
+    align_large = info->allowColorTiling || IS_AVIVO_VARIANT;
 
     /* FIXME: May need to validate line pitch here */
     switch (pScrn->depth / 8) {
-    case 1: if (info->allowColorTiling) dummy = (pScrn->virtualX + 255) & ~255;
-	else dummy = (pScrn->virtualX + 127) & ~127;
+    case 1: pitch_mask = align_large ? 255 : 127;
 	break;
-    case 2: if (info->allowColorTiling) dummy = (pScrn->virtualX + 127) & ~127;
-	else dummy = (pScrn->virtualX + 31) & ~31;
+    case 2: pitch_mask = align_large ? 127 : 31;
 	break;
     case 3:
-    case 4: if (info->allowColorTiling) dummy = (pScrn->virtualX + 63) & ~63;
-	else dummy = (pScrn->virtualX + 15) & ~15;
+    case 4: pitch_mask = align_large ? 63 : 15;
 	break;
     }
+    dummy = (pScrn->virtualX + pitch_mask) & ~pitch_mask;
     pScrn->displayWidth = dummy;
     info->CurrentLayout.displayWidth = pScrn->displayWidth;
 
 }
 
-static DisplayModePtr RADEONTVModes(xf86OutputPtr output)
+static DisplayModePtr
+RADEONTVModes(xf86OutputPtr output)
 {
     DisplayModePtr new  = NULL;
 
@@ -89,6 +92,55 @@ static DisplayModePtr RADEONTVModes(xf86OutputPtr output)
     return new;
 }
 
+static DisplayModePtr
+RADEONATOMTVModes(xf86OutputPtr output)
+{
+    RADEONOutputPrivatePtr radeon_output = output->driver_private;
+    DisplayModePtr  last       = NULL;
+    DisplayModePtr  new        = NULL;
+    DisplayModePtr  first      = NULL;
+    int max_v, i;
+    /* Add some common sizes */
+    int widths[5] = {640, 720, 800, 848, 1024};
+
+    if (radeon_output->tvStd == TV_STD_NTSC ||
+	radeon_output->tvStd == TV_STD_NTSC_J ||
+	radeon_output->tvStd == TV_STD_PAL_M)
+	max_v = 480;
+    else
+	max_v = 600;
+
+    for (i = 0; i < 5; i++) {
+	new = xf86CVTMode(widths[i], max_v, 60.0, FALSE, FALSE);
+
+	new->type       = M_T_DRIVER;
+
+	if (radeon_output->tvStd == TV_STD_NTSC ||
+	    radeon_output->tvStd == TV_STD_NTSC_J ||
+	    radeon_output->tvStd == TV_STD_PAL_M) {
+	    if (widths[i] == 640)
+		new->type |= M_T_PREFERRED;
+	} else {
+	    if (widths[i] == 800)
+		new->type |= M_T_PREFERRED;
+	}
+
+	new->next       = NULL;
+	new->prev       = last;
+
+	if (last) last->next = new;
+	last = new;
+	if (!first) first = new;
+    }
+
+    if (last) {
+	last->next   = NULL; //first;
+	first->prev  = NULL; //last;
+    }
+
+    return first;
+}
+
 /* This is used only when no mode is specified for FP and no ddc is
  * available.  We force it to native mode, if possible.
  */
@@ -97,18 +149,35 @@ static DisplayModePtr RADEONFPNativeMode(xf86OutputPtr output)
     ScrnInfoPtr pScrn = output->scrn;
     RADEONOutputPrivatePtr radeon_output = output->driver_private;
     DisplayModePtr  new   = NULL;
+    char            stmp[32];
 
     if (radeon_output->PanelXRes != 0 &&
 	radeon_output->PanelYRes != 0 &&
 	radeon_output->DotClock != 0) {
 
-	/* Add native panel size */
-	new = xf86CVTMode(radeon_output->PanelXRes, radeon_output->PanelYRes, 60.0, TRUE, FALSE);
+	new             = xnfcalloc(1, sizeof (DisplayModeRec));
+	sprintf(stmp, "%dx%d", radeon_output->PanelXRes, radeon_output->PanelYRes);
+	new->name       = xnfalloc(strlen(stmp) + 1);
+	strcpy(new->name, stmp);
+	new->HDisplay   = radeon_output->PanelXRes;
+	new->VDisplay   = radeon_output->PanelYRes;
 
-	new->type       = M_T_DRIVER | M_T_PREFERRED;
+	new->HTotal     = new->HDisplay + radeon_output->HBlank;
+	new->HSyncStart = new->HDisplay + radeon_output->HOverPlus;
+	new->HSyncEnd   = new->HSyncStart + radeon_output->HSyncWidth;
+	new->VTotal     = new->VDisplay + radeon_output->VBlank;
+	new->VSyncStart = new->VDisplay + radeon_output->VOverPlus;
+	new->VSyncEnd   = new->VSyncStart + radeon_output->VSyncWidth;
 
-	new->next       = NULL;
-	new->prev       = NULL;
+	new->Clock      = radeon_output->DotClock;
+	new->Flags      = 0;
+
+	if (new) {
+	    new->type       = M_T_DRIVER | M_T_PREFERRED;
+
+	    new->next       = NULL;
+	    new->prev       = NULL;
+	}
 
 	xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Added native panel mode: %dx%d\n",
 		   radeon_output->PanelXRes, radeon_output->PanelYRes);
@@ -117,6 +186,65 @@ static DisplayModePtr RADEONFPNativeMode(xf86OutputPtr output)
     return new;
 }
 
+#if defined(__powerpc__)
+/* Apple eMacs need special modes for the internal CRT, e.g.,
+ * Modeline "640x480"    62.12   640  680  752  864  480 481 484  521 +HSync +Vsync
+ * Modeline "800x600"    76.84   800  848  936 1072  600 601 604  640 +HSync +Vsync
+ * Modeline "1024x768"   99.07  1024 1088 1200 1376  768 769 772  809 +HSync +Vsync
+ * Modeline "1152x864"  112.36  1152 1224 1352 1552  864 865 868  905 +HSync +Vsync
+ * Modeline "1280x960"  124.54  1280 1368 1504 1728  960 961 964 1001 +HSync +Vsync
+ */
+static DisplayModePtr RADEONeMacModes(xf86OutputPtr output)
+{
+    ScrnInfoPtr pScrn = output->scrn;
+    DisplayModePtr last=NULL, new=NULL, first=NULL;
+    int i, *modep;
+    static const char *modenames[5] = {
+	"640x480", "800x600", "1024x768", "1152x864", "1280x960"
+    };
+    static int modes[9*5] = {
+	 62120,  640,  680,  752,  864, 480, 481, 484,  521,
+	 76840,  800,  848,  936, 1072, 600, 601, 604,  640,
+	 99070, 1024, 1088, 1200, 1376, 768, 769, 772,  809,
+	112360, 1152, 1224, 1352, 1552, 864, 865, 868,  905,
+	124540, 1280, 1368, 1504, 1728, 960, 961, 964, 1001
+    };
+    modep = modes;
+
+    for (i=0; i<5; i++) {
+	new = xnfcalloc(1, sizeof (DisplayModeRec));
+	if (new) {
+	    new->name       = xnfalloc(strlen(modenames[i]) + 1);
+	    strcpy(new->name, modenames[i]);
+	    new->Clock      = *modep++;
+
+	    new->HDisplay   = *modep++;
+	    new->HSyncStart = *modep++;
+	    new->HSyncEnd   = *modep++;
+	    new->HTotal     = *modep++;
+
+	    new->VDisplay   = *modep++;
+	    new->VSyncStart = *modep++;
+	    new->VSyncEnd   = *modep++;
+	    new->VTotal     = *modep++;
+
+	    new->Flags      = 0;
+	    new->type       = M_T_DRIVER;
+	    if (i==2)
+		new->type |= M_T_PREFERRED;
+	    new->next       = NULL;
+	    new->prev       = last;
+	    if (last) last->next = new;
+	    last = new;
+	    if (!first) first = new;
+	    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Added eMac mode %s\n", modenames[i]);
+	}
+    }
+
+    return first;
+}
+#endif
+
 /* this function is basically a hack to add the screen modes */
 static void RADEONAddScreenModes(xf86OutputPtr output, DisplayModePtr *modeList)
 {
@@ -159,7 +287,7 @@ static void RADEONAddScreenModes(xf86OutputPtr output, DisplayModePtr *modeList)
 	    }
 	}
 
-	new = xf86CVTMode(width, height, 60.0, TRUE, FALSE);
+	new = xf86CVTMode(width, height, 60.0, FALSE, FALSE);
 
 	new->type      |= M_T_USERDEF;
 
@@ -192,37 +320,53 @@ DisplayModePtr
 RADEONProbeOutputModes(xf86OutputPtr output)
 {
     RADEONOutputPrivatePtr radeon_output = output->driver_private;
-    xf86MonPtr		    edid_mon;
+    ScrnInfoPtr pScrn = output->scrn;
+    RADEONInfoPtr info = RADEONPTR(pScrn);
     DisplayModePtr	    modes = NULL;
+    AtomBiosArgRec atomBiosArg;
+    AtomBiosResult atomBiosResult;
 
     ErrorF("in RADEONProbeOutputModes\n");
 
     if (output->status == XF86OutputStatusConnected) {
-	if (radeon_output->type == OUTPUT_DVI || radeon_output->type == OUTPUT_VGA) {
-	    edid_mon = xf86OutputGetEDID (output, radeon_output->pI2CBus);
-	    xf86OutputSetEDID (output, edid_mon);
-
-	    modes = xf86OutputGetEDIDModes (output);
-	    return modes;
-	}
-	if (radeon_output->type == OUTPUT_STV || radeon_output->type == OUTPUT_CTV) {
-	    modes = RADEONTVModes(output);
-	    return modes;
-	}
-	if (radeon_output->type == OUTPUT_LVDS) {
-	    /* okay we got DDC info */
-	    if (output->MonInfo) {
-		edid_mon = xf86OutputGetEDID (output, radeon_output->pI2CBus);
-		xf86OutputSetEDID (output, edid_mon);
-
+	if (OUTPUT_IS_TV) {
+	    if (IS_AVIVO_VARIANT)
+		modes = RADEONATOMTVModes(output);
+	    else
+		modes = RADEONTVModes(output);
+	} else if (radeon_output->type == OUTPUT_CV) {
+	    atomBiosResult = RHDAtomBiosFunc(pScrn->scrnIndex, info->atomBIOS,
+					     ATOMBIOS_GET_CV_MODES, &atomBiosArg);
+	    if (atomBiosResult == ATOM_SUCCESS) {
+		modes = atomBiosArg.modes;
+	    }
+	} else {
+	    if (output->MonInfo)
 		modes = xf86OutputGetEDIDModes (output);
-	    }
+#if defined(__powerpc__)
+	    if ((info->MacModel == RADEON_MAC_EMAC) &&
+		(radeon_output->DACType == DAC_PRIMARY) &&
+		(modes == NULL))
+		modes = RADEONeMacModes(output);
+#endif
 	    if (modes == NULL) {
-		modes = RADEONFPNativeMode(output);
-		/* add the screen modes */
-		RADEONAddScreenModes(output, &modes);
+		if ((radeon_output->type == OUTPUT_LVDS) && info->IsAtomBios) {
+		    atomBiosResult = RHDAtomBiosFunc(pScrn->scrnIndex,
+						     info->atomBIOS,
+						     ATOMBIOS_GET_PANEL_EDID, &atomBiosArg);
+		    if (atomBiosResult == ATOM_SUCCESS) {
+			output->MonInfo = xf86InterpretEDID(pScrn->scrnIndex,
+							    atomBiosArg.EDIDBlock);
+			modes = xf86OutputGetEDIDModes(output);
+		    }
+		}
+		if (modes == NULL) {
+		    if (radeon_output->type == OUTPUT_LVDS)
+			modes = RADEONFPNativeMode(output);
+		    /* add the screen modes */
+		    RADEONAddScreenModes(output, &modes);
+		}
 	    }
-	    return modes;
 	}
     }
 
